/*
   ┏━━━┓╱╱╱╱╱┏┓╱╱┏┓╱╱╱╱╱┏┓╱╱
   ┃┏━┓┃╱╱╱╱╱┃┃╱╱┃┃╱╱╱╱╱┃┃╱╱╱
   ┃┗━┛┣┓┏┳━━┫┃┏┓┃┃╱╱┏━━┫┗━┳━━┓
   ┃┏━━┫┃┃┃┏┓┃┗┛┃┃┃╱┏┫┏┓┃┏┓┃━━┫
   ┃┃╱╱┃┗┛┃┃┃┃┏┓┃┃┗━┛┃┏┓┃┗┛┣━━┃
   ┗┛╱╱┗━━┻┛┗┻┛┗┛┗━━━┻┛┗┻━━┻━━┛
    ━━━━━━━━━━━━━━━━━━━━━━━━━━

    Copyright (c) 2022 Punk Labs LLC

    This section is part of OneTrick CRYPTID

    OneTrick CRYPTID is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the Free
    Software Foundation, either version 3 of the License, or (at your option)
    any later version.

    OneTrick CRYPTID is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
    more details.

    You should have received a copy of the GNU General Public License along with
    OneTrick CRYPTID.  If not, see <http://www.gnu.org/licenses/>.
*/

use std::cell::RefCell;
//use std::rc::Rc;
use std::sync::atomic::{AtomicBool, Ordering};
use std::sync::Arc;

#[cfg(feature = "profile")]
use std::sync::atomic::AtomicI32;

use nih_plug::prelude::*;


#[cfg(feature = "egui")]
use nih_plug_egui::{
    create_egui_editor,
    egui::{
        epaint::{
            Shadow,
        },
        style,
        Margin,
        vec2,
        //Align,
        Align2,
        Area,
        CentralPanel,
        Color32,
        Context,
        FontDefinitions,
        FontFamily,
        FontId,
        Frame,
        Id,
        //LayerId,
        //Layout,
        Order,
        Pos2,
        Rect,
        Rounding,
        Sense,
        Stroke,
        TopBottomPanel,
        Vec2,
    },
    EguiState,
};
// #[cfg(feature = "egui")]
// use egui_extras::{
//     image::RetainedImage,
// };

mod dsp;
use onetrick::prelude::*;

//256 is 187.5Hz at 48kHz
//512 is 94Hz at 48kHz
const MAX_BLOCK_SIZE: usize = 1024;
const NUM_CHANNELS: u32 = 2;
#[cfg(feature = "multi_out")]
const AUX_BUS_COUNT: usize = 11;


// Kick does well tuned to the key of the song
// We'll just assume C2 or A1
const KITNOTE_KICK: u32 = MidiNote::A1 as u32; //C2 or A1
    
// Snare/Rimshot fundamental should be G-ish, Top+Bottom should be Perfect Fourth/Fifth...
// D sounds better with C2 Kick...
const KITNOTE_SNARE: u32 = MidiNote::D3 as u32; //G3 or D3
    
const KITNOTE_RIMSHOT: u32 = MidiNote::D3 as u32;
const KITNOTE_CLAVES: u32 = MidiNote::E3 as u32;
    
const KITNOTE_CLAP: u32 = MidiNote::Ds3 as u32; //Random, what it was already set to
    
// Toms work well with Perfect Fourths: D3 [G3] C3... -5, 0, +5
// Alternatively, a Major Chord works: C3 [A3] F3... -4, 0, +3
const KITNOTE_TOM: u32 = MidiNote::G3 as u32; // Perfect Fourths
//const KITNOTE_TOM: u32 = MidiNote::A3 as u32; // Major Chord
    
const KITNOTE_CYMBAL: u32 = MidiNote::C3 as u32; // Gongs seem to be tuned to C
    
const KITNOTE_COWBELL: u32 = MidiNote::F5 as u32;
const KITNOTE_WHISTLE: u32 = MidiNote::G3 as u32;


// NOTE: Should probably use an array of &dyn OneTrickGeneral, but couldn't make it work
macro_rules! for_each_voice_dsp {
    ($this:ident, $expression:expr) => {{
        $expression(&mut *$this.dsp_kick.borrow_mut());
        $expression(&mut *$this.dsp_snare.borrow_mut());
        $expression(&mut *$this.dsp_rimshot1.borrow_mut());
        $expression(&mut *$this.dsp_rimshot2.borrow_mut());
        $expression(&mut *$this.dsp_clap1.borrow_mut());
        $expression(&mut *$this.dsp_clap2.borrow_mut());
        $expression(&mut *$this.dsp_lowtom1.borrow_mut());
        $expression(&mut *$this.dsp_lowtom2.borrow_mut());
        $expression(&mut *$this.dsp_lowtom3.borrow_mut());
        $expression(&mut *$this.dsp_lowtom4.borrow_mut());
        $expression(&mut *$this.dsp_midtom1.borrow_mut());
        $expression(&mut *$this.dsp_midtom2.borrow_mut());
        $expression(&mut *$this.dsp_midtom3.borrow_mut());
        $expression(&mut *$this.dsp_midtom4.borrow_mut());
        $expression(&mut *$this.dsp_hitom1.borrow_mut());
        $expression(&mut *$this.dsp_hitom2.borrow_mut());
        $expression(&mut *$this.dsp_hitom3.borrow_mut());
        $expression(&mut *$this.dsp_hitom4.borrow_mut());
        $expression(&mut *$this.dsp_hhclosed.borrow_mut());
        $expression(&mut *$this.dsp_hhopen.borrow_mut());
        $expression(&mut *$this.dsp_cymbal1.borrow_mut());
        $expression(&mut *$this.dsp_cymbal2.borrow_mut());
        $expression(&mut *$this.dsp_cymbal3.borrow_mut());
        $expression(&mut *$this.dsp_cowbell1.borrow_mut());
        $expression(&mut *$this.dsp_cowbell2.borrow_mut());
        $expression(&mut *$this.dsp_cowbell3.borrow_mut());
    }};
}

#[cfg(feature = "multi_out")]
macro_rules! for_each_voice_dsp_grouped {
    ($this:ident, $expression:expr) => {{
        $expression(&[&mut *$this.dsp_kick.borrow_mut()]);
        $expression(&[&mut *$this.dsp_snare.borrow_mut()]);
        $expression(&[
            &mut *$this.dsp_rimshot1.borrow_mut(),
            &mut *$this.dsp_rimshot2.borrow_mut(),
        ]);
        $expression(&[
            &mut *$this.dsp_clap1.borrow_mut(),
            &mut *$this.dsp_clap2.borrow_mut(),
        ]);
        $expression(&[
            &mut *$this.dsp_lowtom1.borrow_mut(),
            &mut *$this.dsp_lowtom2.borrow_mut(),
            &mut *$this.dsp_lowtom3.borrow_mut(),
            &mut *$this.dsp_lowtom4.borrow_mut(),
        ]);
        $expression(&[
            &mut *$this.dsp_midtom1.borrow_mut(),
            &mut *$this.dsp_midtom2.borrow_mut(),
            &mut *$this.dsp_midtom3.borrow_mut(),
            &mut *$this.dsp_midtom4.borrow_mut(),
        ]);
        $expression(&[
            &mut *$this.dsp_hitom1.borrow_mut(),
            &mut *$this.dsp_hitom2.borrow_mut(),
            &mut *$this.dsp_hitom3.borrow_mut(),
            &mut *$this.dsp_hitom4.borrow_mut(),
        ]);
        $expression(&[&mut *$this.dsp_hhclosed.borrow_mut()]);
        $expression(&[&mut *$this.dsp_hhopen.borrow_mut()]);
        $expression(&[
            &mut *$this.dsp_cymbal1.borrow_mut(),
            &mut *$this.dsp_cymbal2.borrow_mut(),
            &mut *$this.dsp_cymbal3.borrow_mut(),
        ]);
        $expression(&[
            &mut *$this.dsp_cowbell1.borrow_mut(),
            &mut *$this.dsp_cowbell2.borrow_mut(),
            &mut *$this.dsp_cowbell3.borrow_mut(),
        ]);
    }};
}
macro_rules! for_each_dsp {
    ($this:ident, $expression:expr) => {{
        for_each_voice_dsp!($this, $expression);
        $expression(&mut *$this.dsp_output.borrow_mut());
    }};
}

pub struct OneTrickPlugin {
    params: Arc<OneTrickPluginParams>,

    accum_buffer: ResizableBuffer,

    dsp_output: RefCell<OneTrickDSP<dsp::modules::DSP_Output>>,
    dsp_kick: RefCell<OneTrickDSP<dsp::modules::DSP_Kick>>,
    dsp_snare: RefCell<OneTrickDSP<dsp::modules::DSP_Snare>>,
    dsp_rimshot1: RefCell<OneTrickDSP<dsp::modules::DSP_Rimshot1>>,
    dsp_rimshot2: RefCell<OneTrickDSP<dsp::modules::DSP_Rimshot2>>,
    dsp_clap1: RefCell<OneTrickDSP<dsp::modules::DSP_Clap1>>,
    dsp_clap2: RefCell<OneTrickDSP<dsp::modules::DSP_Clap2>>,
    dsp_lowtom1: RefCell<OneTrickDSP<dsp::modules::DSP_Tom1>>,
    dsp_lowtom2: RefCell<OneTrickDSP<dsp::modules::DSP_Tom2>>,
    dsp_lowtom3: RefCell<OneTrickDSP<dsp::modules::DSP_Tom3>>,
    dsp_lowtom4: RefCell<OneTrickDSP<dsp::modules::DSP_Tom4>>,
    dsp_midtom1: RefCell<OneTrickDSP<dsp::modules::DSP_Tom1>>,
    dsp_midtom2: RefCell<OneTrickDSP<dsp::modules::DSP_Tom2>>,
    dsp_midtom3: RefCell<OneTrickDSP<dsp::modules::DSP_Tom3>>,
    dsp_midtom4: RefCell<OneTrickDSP<dsp::modules::DSP_Tom4>>,
    dsp_hitom1: RefCell<OneTrickDSP<dsp::modules::DSP_Tom1>>,
    dsp_hitom2: RefCell<OneTrickDSP<dsp::modules::DSP_Tom2>>,
    dsp_hitom3: RefCell<OneTrickDSP<dsp::modules::DSP_Tom3>>,
    dsp_hitom4: RefCell<OneTrickDSP<dsp::modules::DSP_Tom4>>,
    dsp_hhclosed: RefCell<OneTrickDSP<dsp::modules::DSP_HihatClosed>>,
    dsp_hhopen: RefCell<OneTrickDSP<dsp::modules::DSP_HihatOpen>>,
    dsp_cymbal1: RefCell<OneTrickDSP<dsp::modules::DSP_Cymbal1>>,
    dsp_cymbal2: RefCell<OneTrickDSP<dsp::modules::DSP_Cymbal2>>,
    dsp_cymbal3: RefCell<OneTrickDSP<dsp::modules::DSP_Cymbal3>>,
    dsp_cowbell1: RefCell<OneTrickDSP<dsp::modules::DSP_Cowbell1>>,
    dsp_cowbell2: RefCell<OneTrickDSP<dsp::modules::DSP_Cowbell2>>,
    dsp_cowbell3: RefCell<OneTrickDSP<dsp::modules::DSP_Cowbell3>>,
    //dsp: [dyn OneTrickDSPGeneral],
    param_global_sensitivity: usize,
    param_rimshot_voice: usize,
    param_clap_voice: usize,
    param_lowtom_alt_transpose: usize,
    param_lowtom_voice: usize,
    param_midtom_alt_transpose: usize,
    param_midtom_voice: usize,
    param_hitom_alt_transpose: usize,
    param_hitom_voice: usize,
    param_cymbal_voice: usize,
    param_cowbell_alt_transpose: usize,
    param_cowbell_voice: usize,

    kick_indicator: Arc<AtomicBool>,
    snare_indicator: Arc<AtomicBool>,
    rimshot_indicator: Arc<AtomicBool>,
    clap_indicator: Arc<AtomicBool>,
    lowtom_indicator: Arc<AtomicBool>,
    midtom_indicator: Arc<AtomicBool>,
    hitom_indicator: Arc<AtomicBool>,
    hhclosed_indicator: Arc<AtomicBool>,
    hhopen_indicator: Arc<AtomicBool>,
    cymbal_indicator: Arc<AtomicBool>,
    cowbell_indicator: Arc<AtomicBool>,

    #[cfg(feature = "profile")]
    dsp_cost: Arc<AtomicI32>,
    sample_rate: usize,
}

impl Default for OneTrickPlugin {
    fn default() -> Self {
        /*
        {
            use faust_types::FaustDsp;
            let explosion = Box::new(dsp::modules::DSP_Output::new());
        }
        */

        /* Generate Params?
            let params = OneTrickPluginParams::default();
            // Maybe use indices to reference subgroups?... Group names can be used by NIH-Plug
            params.append_dsp(&dsp_kick, 0, "Kick");
            params.append_dsp(&dsp_snare, 1, "Snare");
            ...
            params.append_dsp(&dsp_output, 11, "");

            //Later when we need to update params:
            params.apply_to_dsp(&mut dsp_kick, 0);
            params.apply_to_dsp(&mut dsp_snare, 1);
            ...
            params.apply_to_dsp(&mut dsp_output, 11);
        */

        let mut params = OneTrickPluginParams::default();
        #[cfg(feature = "egui")]
        {
            params.editor_state = EguiState::from_size(1090, 630 + 95 + 30+10);
        }

        let mut dsp_output = OneTrickDSP::new();
        let mut dsp_kick = OneTrickDSP::new();
        let mut dsp_snare = OneTrickDSP::new();
        let mut dsp_rimshot1 = OneTrickDSP::new();
        let mut dsp_rimshot2 = OneTrickDSP::new();
        let mut dsp_clap1 = OneTrickDSP::new();
        let mut dsp_clap2 = OneTrickDSP::new();
        let mut dsp_lowtom1 = OneTrickDSP::new();
        let mut dsp_lowtom2 = OneTrickDSP::new();
        let mut dsp_lowtom3 = OneTrickDSP::new();
        let mut dsp_lowtom4 = OneTrickDSP::new();
        let mut dsp_midtom1 = OneTrickDSP::new();
        let mut dsp_midtom2 = OneTrickDSP::new();
        let mut dsp_midtom3 = OneTrickDSP::new();
        let mut dsp_midtom4 = OneTrickDSP::new();
        let mut dsp_hitom1 = OneTrickDSP::new();
        let mut dsp_hitom2 = OneTrickDSP::new();
        let mut dsp_hitom3 = OneTrickDSP::new();
        let mut dsp_hitom4 = OneTrickDSP::new();
        let mut dsp_hhclosed = OneTrickDSP::new();
        let mut dsp_hhopen = OneTrickDSP::new();
        let mut dsp_cymbal1 = OneTrickDSP::new();
        let mut dsp_cymbal2 = OneTrickDSP::new();
        let mut dsp_cymbal3 = OneTrickDSP::new();
        let mut dsp_cowbell1 = OneTrickDSP::new();
        let mut dsp_cowbell2 = OneTrickDSP::new();
        let mut dsp_cowbell3 = OneTrickDSP::new();

        // Override deault reverb for Kick:
        for p in dsp_kick.params_mut().params_mut() {
            if p.name() == "Voice_Reverb" {
                p.set_init(0.0);
            }
        }

        // Override deault tuning for Toms:
        for p in dsp_lowtom1.params_mut().params_mut() {
            if p.name() == "Voice_Transpose" {
                p.set_init(-5.0);
            }
        }
        for p in dsp_hitom1.params_mut().params_mut() {
            if p.name() == "Voice_Transpose" {
                p.set_init(5.0);
            }
        }

        params.append_dsp(&mut dsp_output, "Global");
        let param_global_sensitivity = params.append_float(
            FloatParam::new(
                "Global Sensitivity",
                100.0,
                FloatRange::Linear {
                    min: 0.0,
                    max: 100.0,
                },
            )
            .with_unit("%")
            .with_value_to_string(Arc::new(|v| {
                format!("{value:.precision$}", precision = 1, value = v)
            })),
            "Global",
        );

        params.append_dsp(&mut dsp_kick, "Kick");
        params.append_dsp(&mut dsp_snare, "Snare");

        params.append_dsp(&mut dsp_rimshot1, "Rimshot");
        params.append_dsp(&mut dsp_rimshot2, "Rimshot"); // Shares "Rimshot"
        let param_rimshot_voice = params.append_int(
            IntParam::new("Rimshot Voice", 0, IntRange::Linear { min: 0, max: 1 })
                .with_value_to_string(Arc::new(|v| {
                    match v {
                        0 => "Rimshot",
                        1 => "Claves",
                        _ => "?",
                    }
                    .to_string()
                }))
                .with_string_to_value(Arc::new(|s| {
                    match s {
                        "Rimshot" => Some(0),
                        "Claves" => Some(1),
                        _ => None
                    }
                })),
            "Rimshot",
        );

        params.append_dsp(&mut dsp_clap1, "Clap");
        params.append_dsp(&mut dsp_clap2, "Clap"); // Shares "Clap"
        let param_clap_voice = params.append_int(
            IntParam::new("Clap Voice", 0, IntRange::Linear { min: 0, max: 1 })
                .with_value_to_string(Arc::new(|v| {
                    match v {
                        0 => "Clap",
                        1 => "Maraca",
                        _ => "?",
                    }
                    .to_string()
                }))
                .with_string_to_value(Arc::new(|s| {
                    match s {
                        "Clap" => Some(0),
                        "Maraca" => Some(1),
                        _ => None
                    }
                })),
            "Clap",
        );

        params.append_dsp(&mut dsp_lowtom1, "Low Tom");
        params.append_dsp(&mut dsp_lowtom2, "Low Tom"); // Shares "Low Tom"
        params.append_dsp(&mut dsp_lowtom3, "Low Tom"); // Shares "Low Tom"
        params.append_dsp(&mut dsp_lowtom4, "Low Tom"); // Shares "Low Tom"
        let param_lowtom_alt_transpose = params.append_float(
            FloatParam::new(
                "Low Tom Alt Transpose",
                2.0,
                FloatRange::Linear {
                    min: 0.0,
                    max: 12.0,
                },
            )
            .with_unit("st")
            .with_value_to_string(Arc::new(|v| {
                format!("{value:.precision$}", precision = 1, value = v)
            })),
            "Low Tom",
        );
        let param_lowtom_voice = params.append_int(
            IntParam::new("Low Tom Voice", 0, IntRange::Linear { min: 0, max: 3 })
                .with_value_to_string(Arc::new(|v| {
                    match v {
                        0 => "Tom",
                        1 => "Conga",
                        2 => "Bongo",
                        3 => "Log",
                        _ => "?",
                    }
                    .to_string()
                }))
                .with_string_to_value(Arc::new(|s| {
                    match s {
                        "Tom" => Some(0),
                        "Conga" => Some(1),
                        "Bongo" => Some(2),
                        "Log" => Some(3),
                        _ => None
                    }
                })),
            "Low Tom",
        );
        params.append_dsp(&mut dsp_midtom1, "Mid Tom");
        params.append_dsp(&mut dsp_midtom2, "Mid Tom"); // Shares "Mid Tom"
        params.append_dsp(&mut dsp_midtom3, "Mid Tom"); // Shares "Mid Tom"
        params.append_dsp(&mut dsp_midtom4, "Mid Tom"); // Shares "Mid Tom"
        let param_midtom_alt_transpose = params.append_float(
            FloatParam::new(
                "Mid Tom Alt Transpose",
                2.0,
                FloatRange::Linear {
                    min: 0.0,
                    max: 12.0,
                },
            )
            .with_unit("st")
            .with_value_to_string(Arc::new(|v| {
                format!("{value:.precision$}", precision = 1, value = v)
            })),
            "Mid Tom",
        );
        let param_midtom_voice = params.append_int(
            IntParam::new("Mid Tom Voice", 0, IntRange::Linear { min: 0, max: 3 })
                .with_value_to_string(Arc::new(|v| {
                    match v {
                        0 => "Tom",
                        1 => "Conga",
                        2 => "Bongo",
                        3 => "Log",
                        _ => "?",
                    }
                    .to_string()
                }))
                .with_string_to_value(Arc::new(|s| {
                    match s {
                        "Tom" => Some(0),
                        "Conga" => Some(1),
                        "Bongo" => Some(2),
                        "Log" => Some(3),
                        _ => None
                    }
                })),
            "Mid Tom",
        );
        params.append_dsp(&mut dsp_hitom1, "High Tom");
        params.append_dsp(&mut dsp_hitom2, "High Tom"); // Shares "High Tom"
        params.append_dsp(&mut dsp_hitom3, "High Tom"); // Shares "High Tom"
        params.append_dsp(&mut dsp_hitom4, "High Tom"); // Shares "High Tom"
        let param_hitom_alt_transpose = params.append_float(
            FloatParam::new(
                "High Tom Alt Transpose",
                2.0,
                FloatRange::Linear {
                    min: 0.0,
                    max: 12.0,
                },
            )
            .with_unit("st")
            .with_value_to_string(Arc::new(|v| {
                format!("{value:.precision$}", precision = 1, value = v)
            })),
            "High Tom",
        );
        let param_hitom_voice = params.append_int(
            IntParam::new("High Tom Voice", 0, IntRange::Linear { min: 0, max: 3 })
                .with_value_to_string(Arc::new(|v| {
                    match v {
                        0 => "Tom",
                        1 => "Conga",
                        2 => "Bongo",
                        3 => "Log",
                        _ => "?",
                    }
                    .to_string()
                }))
                .with_string_to_value(Arc::new(|s| {
                    match s {
                        "Tom" => Some(0),
                        "Conga" => Some(1),
                        "Bongo" => Some(2),
                        "Log" => Some(3),
                        _ => None
                    }
                })),
            "High Tom",
        );
        params.append_dsp(&mut dsp_hhclosed, "Hihat Closed");
        params.append_dsp(&mut dsp_hhopen, "Hihat Open");

        params.append_dsp(&mut dsp_cymbal1, "Cymbal");
        params.append_dsp(&mut dsp_cymbal2, "Cymbal"); // Shares "Cymbal"
        params.append_dsp(&mut dsp_cymbal3, "Cymbal"); // Shares "Cymbal"
        let param_cymbal_voice = params.append_int(
            IntParam::new("Cymbal Voice", 0, IntRange::Linear { min: 0, max: 2 })
                .with_value_to_string(Arc::new(|v| {
                    match v {
                        0 => "Cymbal",
                        1 => "Gong",
                        2 => "RevCym",
                        _ => "?",
                    }
                    .to_string()
                }))
                .with_string_to_value(Arc::new(|s| {
                    match s {
                        "Cymbal" => Some(0),
                        "Gong" => Some(1),
                        "RevCym" => Some(2),
                        _ => None
                    }
                })),
            "Cymbal",
        );

        params.append_dsp(&mut dsp_cowbell1, "Cowbell");
        params.append_dsp(&mut dsp_cowbell2, "Cowbell"); // Shares "Cowbell"
        params.append_dsp(&mut dsp_cowbell3, "Cowbell"); // Shares "Cowbell"
        let param_cowbell_alt_transpose = params.append_float(
            FloatParam::new(
                "Cowbell Alt Transpose",
                2.0,
                FloatRange::Linear {
                    min: 0.0,
                    max: 12.0,
                },
            )
            .with_unit("st")
            .with_value_to_string(Arc::new(|v| {
                format!("{value:.precision$}", precision = 1, value = v)
            })),
            "Cowbell Tom",
        );
        let param_cowbell_voice = params.append_int(
            IntParam::new("Cowbell Voice", 0, IntRange::Linear { min: 0, max: 2 })
                .with_value_to_string(Arc::new(|v| {
                    match v {
                        0 => "Cowbell",
                        1 => "W.Block",
                        2 => "Whistle",
                        _ => "?",
                    }
                    .to_string()
                }))
                .with_string_to_value(Arc::new(|s| {
                    match s {
                        "Cowbell" => Some(0),
                        "W.Block" => Some(1),
                        "Whistle" => Some(2),
                        _ => None
                    }
                })),
            "Cowbell",
        );

        Self {
            //params: Arc::new(OneTrickPluginParams::default()),
            params: Arc::new(params),

            accum_buffer: ResizableBuffer::default(),

            dsp_output: RefCell::new(dsp_output),
            dsp_kick: RefCell::new(dsp_kick),
            dsp_snare: RefCell::new(dsp_snare),
            dsp_rimshot1: RefCell::new(dsp_rimshot1),
            dsp_rimshot2: RefCell::new(dsp_rimshot2),
            dsp_clap1: RefCell::new(dsp_clap1),
            dsp_clap2: RefCell::new(dsp_clap2),
            dsp_lowtom1: RefCell::new(dsp_lowtom1),
            dsp_lowtom2: RefCell::new(dsp_lowtom2),
            dsp_lowtom3: RefCell::new(dsp_lowtom3),
            dsp_lowtom4: RefCell::new(dsp_lowtom4),
            dsp_midtom1: RefCell::new(dsp_midtom1),
            dsp_midtom2: RefCell::new(dsp_midtom2),
            dsp_midtom3: RefCell::new(dsp_midtom3),
            dsp_midtom4: RefCell::new(dsp_midtom4),
            dsp_hitom1: RefCell::new(dsp_hitom1),
            dsp_hitom2: RefCell::new(dsp_hitom2),
            dsp_hitom3: RefCell::new(dsp_hitom3),
            dsp_hitom4: RefCell::new(dsp_hitom4),
            dsp_hhclosed: RefCell::new(dsp_hhclosed),
            dsp_hhopen: RefCell::new(dsp_hhopen),
            dsp_cymbal1: RefCell::new(dsp_cymbal1),
            dsp_cymbal2: RefCell::new(dsp_cymbal2),
            dsp_cymbal3: RefCell::new(dsp_cymbal3),
            dsp_cowbell1: RefCell::new(dsp_cowbell1),
            dsp_cowbell2: RefCell::new(dsp_cowbell2),
            dsp_cowbell3: RefCell::new(dsp_cowbell3),

            param_global_sensitivity,
            param_rimshot_voice,
            param_clap_voice,
            param_lowtom_alt_transpose,
            param_lowtom_voice,
            param_midtom_alt_transpose,
            param_midtom_voice,
            param_hitom_alt_transpose,
            param_hitom_voice,
            param_cymbal_voice,
            param_cowbell_alt_transpose,
            param_cowbell_voice,

            kick_indicator: Arc::new(AtomicBool::new(false)),
            snare_indicator: Arc::new(AtomicBool::new(false)),
            rimshot_indicator: Arc::new(AtomicBool::new(false)),
            clap_indicator: Arc::new(AtomicBool::new(false)),
            lowtom_indicator: Arc::new(AtomicBool::new(false)),
            midtom_indicator: Arc::new(AtomicBool::new(false)),
            hitom_indicator: Arc::new(AtomicBool::new(false)),
            hhclosed_indicator: Arc::new(AtomicBool::new(false)),
            hhopen_indicator: Arc::new(AtomicBool::new(false)),
            cymbal_indicator: Arc::new(AtomicBool::new(false)),
            cowbell_indicator: Arc::new(AtomicBool::new(false)),

            #[cfg(feature = "profile")]
            dsp_cost: Arc::new(AtomicI32::new(0)),

            sample_rate: 48000,
        }
    }
}

impl OneTrickPlugin {
    #[cfg(feature = "egui")]
    fn setup_egui_fonts(ctx: &Context) {
        // Start with the default fonts (we will be adding to them rather than replacing them).
        let mut fonts = FontDefinitions::default();

        BasicFonts::add_fonts(&mut fonts);
        Icons::add_fonts(&mut fonts);

        let mut style = style::Style::default();

        // Setup defaults for using generic controls:
        let font_id = FontId::new(17.0, FontFamily::Proportional);
        style.override_font_id = Some(font_id);
        //style.override_text_style = Some(TextStyle::Body);

        //style.widgets.noninteractive.bg_stroke = 
        //    Stroke::new(2.0, palette.grey().alpha(0.666).to_color32());

        // Set the base Style
        style.visuals = style::Visuals::dark();

        // Tooltips and Popup Windows:
        style.visuals.window_fill = Color32::DARK_GRAY;
        style.visuals.window_stroke = Stroke::default();
        style.visuals.popup_shadow = Shadow {offset: Vec2::new(0.0, 0.0), blur: 6.0, spread: 0.0, color: Color32::from_black_alpha(32)};
        style.visuals.override_text_color = Some(Color32::WHITE);
        
        style.visuals.selection.stroke = Stroke::new(2.0, Color32::WHITE);
        ctx.set_style(style);

        // Tell egui to use these fonts:
        ctx.set_fonts(fonts);
    }

    fn map_midi_note(note: u8, channel: u8) -> Option<GeneralMidiDrums> {
        match channel {
            1  => Some(GeneralMidiDrums::BassDrum),
            2  => Some(GeneralMidiDrums::AcousticSnare),
            3  => Some(GeneralMidiDrums::SideStick),
            4  => Some(GeneralMidiDrums::HandClap),
            5  => Some(GeneralMidiDrums::LowFloorTom),
            6  => Some(GeneralMidiDrums::LowTom),
            7  => Some(GeneralMidiDrums::HighMidTom),
            8  => Some(GeneralMidiDrums::ClosedHihat),
            10  => Some(GeneralMidiDrums::OpenHihat),
            11 => Some(GeneralMidiDrums::CrashCymbal),
            12 => Some(GeneralMidiDrums::Cowbell),
            _ => GeneralMidiDrums::try_from(note).ok()
        }        
    }
    fn map_midi_chromatic(note: u8, channel: u8) -> Option<f32> {
        match channel {
            1  => Some(note as f32),
            2  => Some(note as f32),
            3  => Some(note as f32),
            4  => Some(note as f32),
            5  => Some(note as f32),
            6  => Some(note as f32),
            7  => Some(note as f32),
            8  => Some(note as f32),
            10  => Some(note as f32),
            11 => Some(note as f32),
            12 => Some(note as f32),
            _ => None
        }        
    }
}

impl Plugin for OneTrickPlugin {
    const NAME: &'static str = "OneTrick CRYPTID";
    const VENDOR: &'static str = "Punk Labs LLC";
    const URL: &'static str = env!("CARGO_PKG_HOMEPAGE");
    const EMAIL: &'static str = "contact@punklabs.com";

    const VERSION: &'static str = env!("CARGO_PKG_VERSION");

    const AUDIO_IO_LAYOUTS: &'static [AudioIOLayout] = &[
        #[cfg(feature = "multi_out")]
        AudioIOLayout {
            // Multi-Out Stereo
            main_input_channels: NonZeroU32::new(0),
            main_output_channels: NonZeroU32::new(NUM_CHANNELS),
            aux_input_ports: &[],
            // Two to five of these ports will be used at a time
            aux_output_ports: &[new_nonzero_u32(NUM_CHANNELS); AUX_BUS_COUNT],

            names: PortNames {
                layout: Some("Multi-Out Stereo"),

                main_input: None,
                main_output: Some("Output"),
                aux_inputs: &[],
                aux_outputs: &[
                    "Kick",
                    "Snare",
                    "Rimshot",
                    "Clap",
                    "Low Tom",
                    "Mid Tom",
                    "Hi Tom",
                    "Hihat Closed",
                    "Hihat Open",
                    "Cymbal",
                    "Cowbell",
                ],
            }
        },
        AudioIOLayout {
            // Stereo
            main_input_channels: NonZeroU32::new(0),
            main_output_channels: NonZeroU32::new(NUM_CHANNELS),
            names: PortNames {
                layout: Some("Stereo"),

                main_input: None,
                main_output: Some("Output"),
                aux_inputs: &[],
                aux_outputs: &[],
            },
            ..AudioIOLayout::const_default()
        },
    ];

    const MIDI_INPUT: MidiConfig = MidiConfig::Basic; //Basic or MidiCCs
    const MIDI_OUTPUT: MidiConfig = MidiConfig::None;

    const SAMPLE_ACCURATE_AUTOMATION: bool = false; // Too bad this doesn't split on Midi NoteOn events

    type SysExMessage = ();
    type BackgroundTask = ();

    fn params(&self) -> Arc<dyn Params> {
        self.params.clone()
    }

    fn initialize(
        &mut self,
        _audio_io_layout: &AudioIOLayout,
        buffer_config: &BufferConfig,
        _context: &mut impl InitContext<Self>,
    ) -> bool {
        // Resize buffers and perform other potentially expensive initialization operations here.
        // The `reset()` function is always called right after this function. You can remove this
        // function if you do not need it.

        self.accum_buffer.resize(3, MAX_BLOCK_SIZE);

        let sample_rate = buffer_config.sample_rate as usize;
        //println!("Sample Rate: {}", sample_rate);
        self.sample_rate = sample_rate;
        const SILENCE_MS: u32 = 100;
        for_each_dsp!(self, &mut |dsp: &mut dyn OneTrickDSPGeneral| {
            dsp.initialize(sample_rate)
                .resize_buffer(MAX_BLOCK_SIZE)
                .track_silence(SILENCE_MS);
        });

        true
    }

    fn reset(&mut self) {
        // Reset buffers and envelopes here. This can be called from the audio thread and may not
        // allocate. You can remove this function if you do not need it.

        //self.dsp_output.instance_clear();
        //self.dsp_kick.borrow_mut().instance_clear();
        //self.dsp_cowbell.instance_clear();
    }
    
    fn process(
        &mut self,
        buffer: &mut Buffer,
        aux: &mut AuxiliaryBuffers,
        context: &mut impl ProcessContext<Self>,
    ) -> ProcessStatus {
        #[cfg(feature = "profile")]
        let dsp_timer = std::time::Instant::now();

        ProcessEx::process_split_notes(
            buffer,
            aux,
            MAX_BLOCK_SIZE,
            context,
            |event| {
                // MIDI
                match event {
                    NoteEvent::NoteOn {
                        timing: _,
                        voice_id: _,
                        channel,
                        note,
                        velocity,
                    } => {
                            // Allow overriding note data using channels
                        if let Some(midi_drum_note) = Self::map_midi_note(note, channel) {
                            let chromatic = Self::map_midi_chromatic(note, channel);
                            let velocity = if let Some(p) =
                                self.params.param_float_at(self.param_global_sensitivity)
                            {
                                let sensitivity = p.value() * 0.01;
                                velocity * sensitivity + (1.0 - sensitivity) * (80.0/127.0)
                            } else {
                                velocity
                            };

                            match midi_drum_note {
                                GeneralMidiDrums::BassDrum | GeneralMidiDrums::AcousticBassDrum => {
                                    let alt: f32 = match midi_drum_note {
                                        GeneralMidiDrums::AcousticBassDrum => 0.0,//alt_transpose,
                                        _ => 0.0,
                                    };
                                    self.kick_indicator.store(true, Ordering::Relaxed);
                                    self.dsp_output.borrow_mut().wake_up();
                                    self.dsp_kick.borrow_mut().note_on(chromatic.unwrap_or(KITNOTE_KICK as f32 + alt), velocity);
                                }
                                GeneralMidiDrums::AcousticSnare
                                | GeneralMidiDrums::ElectricSnare => {
                                    let alt: f32 = match midi_drum_note {
                                        GeneralMidiDrums::ElectricSnare => 0.0,//alt_transpose,
                                        _ => 0.0,
                                    };
                                    self.snare_indicator.store(true, Ordering::Relaxed);
                                    self.dsp_output.borrow_mut().wake_up();
                                    self.dsp_snare.borrow_mut().note_on(chromatic.unwrap_or(KITNOTE_SNARE as f32 + alt), velocity);
                                }
                                GeneralMidiDrums::SideStick => {
                                    self.rimshot_indicator.store(true, Ordering::Relaxed);
                                    self.dsp_output.borrow_mut().wake_up();
                                    let voice_index = if let Some(p) =
                                        self.params.param_int_at(self.param_rimshot_voice)
                                    {
                                        p.value()
                                    } else {
                                        0
                                    };
                                    match voice_index {
                                        0 => {
                                            self.dsp_rimshot1
                                                .borrow_mut()
                                                .note_on(chromatic.unwrap_or(KITNOTE_RIMSHOT as f32), velocity);
                                            }
                                        1 => {
                                            self.dsp_rimshot2
                                                .borrow_mut()
                                                .note_on(chromatic.unwrap_or(KITNOTE_CLAVES as f32), velocity);
                                        }
                                        _ => {}
                                    }
                                }
                                GeneralMidiDrums::HandClap => {
                                    self.clap_indicator.store(true, Ordering::Relaxed);
                                    self.dsp_output.borrow_mut().wake_up();
                                    let voice_index = if let Some(p) =
                                        self.params.param_int_at(self.param_clap_voice)
                                    {
                                        p.value()
                                    } else {
                                        0
                                    };
                                    match voice_index {
                                        0 => {
                                            self.dsp_clap1
                                                .borrow_mut()
                                                .note_on(chromatic.unwrap_or(KITNOTE_CLAP as f32), velocity);
                                            }
                                        1 => {
                                            self.dsp_clap2
                                                .borrow_mut()
                                                .note_on(chromatic.unwrap_or(KITNOTE_CLAP as f32), velocity);
                                        }
                                        _ => {}
                                    }
                                }
                                GeneralMidiDrums::LowFloorTom | GeneralMidiDrums::HighFloorTom => {
                                    let alt_transpose = if let Some(p) =
                                        self.params.param_float_at(self.param_lowtom_alt_transpose)
                                    {
                                        p.value()
                                    } else {
                                        0.0
                                    };
                                    let alt: f32 = match midi_drum_note {
                                        GeneralMidiDrums::HighFloorTom => alt_transpose,
                                        _ => 0.0,
                                    };
                                    self.lowtom_indicator.store(true, Ordering::Relaxed);
                                    self.dsp_output.borrow_mut().wake_up();
                                    let voice_index = if let Some(p) =
                                        self.params.param_int_at(self.param_lowtom_voice)
                                    {
                                        p.value()
                                    } else {
                                        0
                                    };
                                    match voice_index {
                                        0 => {
                                            self.dsp_lowtom1
                                                .borrow_mut()
                                                .note_on(chromatic.unwrap_or(KITNOTE_TOM as f32 + alt), velocity);
                                        }
                                        1 => {
                                            self.dsp_lowtom2
                                                .borrow_mut()
                                                .note_on(chromatic.unwrap_or(KITNOTE_TOM as f32 + alt), velocity);
                                        }
                                        2 => {
                                            self.dsp_lowtom3
                                                .borrow_mut()
                                                .note_on(chromatic.unwrap_or(KITNOTE_TOM as f32 + alt), velocity);
                                        }
                                        3 => {
                                            self.dsp_lowtom4
                                                .borrow_mut()
                                                .note_on(chromatic.unwrap_or(KITNOTE_TOM as f32 + alt), velocity);
                                        }
                                        _ => {}
                                    }
                                }
                                GeneralMidiDrums::LowTom | GeneralMidiDrums::LowMidTom => {
                                    let alt_transpose = if let Some(p) =
                                        self.params.param_float_at(self.param_midtom_alt_transpose)
                                    {
                                        p.value()
                                    } else {
                                        0.0
                                    };
                                    let alt: f32 = match midi_drum_note {
                                        GeneralMidiDrums::LowMidTom => alt_transpose,
                                        _ => 0.0,
                                    };
                                    self.midtom_indicator.store(true, Ordering::Relaxed);
                                    self.dsp_output.borrow_mut().wake_up();
                                    let voice_index = if let Some(p) =
                                        self.params.param_int_at(self.param_midtom_voice)
                                    {
                                        p.value()
                                    } else {
                                        0
                                    };
                                    match voice_index {
                                        0 => {
                                            self.dsp_midtom1
                                                .borrow_mut()
                                                .note_on(chromatic.unwrap_or(KITNOTE_TOM as f32 + alt), velocity);
                                        }
                                        1 => {
                                            self.dsp_midtom2
                                                .borrow_mut()
                                                .note_on(chromatic.unwrap_or(KITNOTE_TOM as f32 + alt), velocity);
                                        }
                                        2 => {
                                            self.dsp_midtom3
                                                .borrow_mut()
                                                .note_on(chromatic.unwrap_or(KITNOTE_TOM as f32 + alt), velocity);
                                        }
                                        3 => {
                                            self.dsp_midtom4
                                                .borrow_mut()
                                                .note_on(chromatic.unwrap_or(KITNOTE_TOM as f32 + alt), velocity);
                                        }
                                        _ => {}
                                    }
                                }
                                GeneralMidiDrums::HighMidTom | GeneralMidiDrums::HighTom => {
                                    let alt_transpose = if let Some(p) =
                                        self.params.param_float_at(self.param_hitom_alt_transpose)
                                    {
                                        p.value()
                                    } else {
                                        0.0
                                    };
                                    let alt: f32 = match midi_drum_note {
                                        GeneralMidiDrums::HighTom => alt_transpose,
                                        _ => 0.0,
                                    };
                                    self.hitom_indicator.store(true, Ordering::Relaxed);
                                    self.dsp_output.borrow_mut().wake_up();
                                    let voice_index = if let Some(p) =
                                        self.params.param_int_at(self.param_hitom_voice)
                                    {
                                        p.value()
                                    } else {
                                        0
                                    };
                                    match voice_index {
                                        0 => {
                                            self.dsp_hitom1
                                                .borrow_mut()
                                                .note_on(chromatic.unwrap_or(KITNOTE_TOM as f32 + alt), velocity);
                                        }
                                        1 => {
                                            self.dsp_hitom2
                                                .borrow_mut()
                                                .note_on(chromatic.unwrap_or(KITNOTE_TOM as f32 + alt), velocity);
                                        }
                                        2 => {
                                            self.dsp_hitom3
                                                .borrow_mut()
                                                .note_on(chromatic.unwrap_or(KITNOTE_TOM as f32 + alt), velocity);
                                        }
                                        3 => {
                                            self.dsp_hitom4
                                                .borrow_mut()
                                                .note_on(chromatic.unwrap_or(KITNOTE_TOM as f32 + alt), velocity);
                                        }
                                        _ => {}
                                    }
                                }
                                GeneralMidiDrums::ClosedHihat => {
                                    self.hhclosed_indicator.store(true, Ordering::Relaxed);
                                    self.dsp_output.borrow_mut().wake_up();
                                    self.dsp_hhclosed
                                        .borrow_mut()
                                        .note_on(chromatic.unwrap_or(note as f32), velocity);
                                    self.dsp_hhopen.borrow_mut().choke();
                                }
                                GeneralMidiDrums::PedalHihat => {
                                    self.dsp_output.borrow_mut().wake_up();
                                    self.dsp_hhclosed.borrow_mut().choke();
                                    self.dsp_hhopen.borrow_mut().choke();
                                }
                                GeneralMidiDrums::OpenHihat => {
                                    self.hhopen_indicator.store(true, Ordering::Relaxed);
                                    self.dsp_output.borrow_mut().wake_up();
                                    self.dsp_hhopen.borrow_mut().note_on(chromatic.unwrap_or(note as f32), velocity);
                                }
                                GeneralMidiDrums::CrashCymbal => {
                                    self.cymbal_indicator.store(true, Ordering::Relaxed);
                                    self.dsp_output.borrow_mut().wake_up();
                                    let voice_index = if let Some(p) =
                                        self.params.param_int_at(self.param_cymbal_voice)
                                    {
                                        p.value()
                                    } else {
                                        0
                                    };
                                    match voice_index {
                                        0 => {
                                            self.dsp_cymbal1.borrow_mut().note_on(chromatic.unwrap_or(KITNOTE_CYMBAL as f32), velocity);
                                        }
                                        1 => {
                                            self.dsp_cymbal2.borrow_mut().note_on(chromatic.unwrap_or(KITNOTE_CYMBAL as f32), velocity);
                                        }
                                        2 => {
                                            self.dsp_cymbal3.borrow_mut().note_on(chromatic.unwrap_or(KITNOTE_CYMBAL as f32), velocity);
                                        }
                                        _ => {}
                                    }
                                }
                                GeneralMidiDrums::RideCymbal => {
                                    self.dsp_output.borrow_mut().wake_up();
                                    self.dsp_cymbal1.borrow_mut().choke();
                                    self.dsp_cymbal2.borrow_mut().choke();
                                    self.dsp_cymbal3.borrow_mut().choke();
                                }
                                GeneralMidiDrums::Cowbell | GeneralMidiDrums::Tambourine => {
                                    let alt_transpose = if let Some(p) =
                                        self.params.param_float_at(self.param_cowbell_alt_transpose)
                                    {
                                        p.value()
                                    } else {
                                        0.0
                                    };
                                    let alt: f32 = match midi_drum_note {
                                        GeneralMidiDrums::Cowbell => alt_transpose,
                                        _ => 0.0,
                                    };
                                    self.cowbell_indicator.store(true, Ordering::Relaxed);
                                    self.dsp_output.borrow_mut().wake_up();
                                    let voice_index = if let Some(p) =
                                        self.params.param_int_at(self.param_cowbell_voice)
                                    {
                                        p.value()
                                    } else {
                                        0
                                    };

                                    match voice_index {
                                        0 => {
                                            self.dsp_cowbell1
                                                .borrow_mut()
                                                .note_on(chromatic.unwrap_or((KITNOTE_COWBELL as f32 + alt) as f32), velocity);
                                        }
                                        1 => {
                                            self.dsp_cowbell2
                                                .borrow_mut()
                                                .note_on(chromatic.unwrap_or((KITNOTE_COWBELL as f32 + alt) as f32), velocity);
                                        }
                                        2 => {
                                            self.dsp_cowbell3
                                                .borrow_mut()
                                                .note_on(chromatic.unwrap_or((KITNOTE_WHISTLE as f32 + alt) as f32), velocity);
                                        }
                                        _ => {}
                                    }
                                }
                                #[cfg(feature = "profile")]
                                GeneralMidiDrums::RideCymbal2 => {
                                    //B to profile Output DSP only
                                    self.dsp_output.borrow_mut().track_silence(1000); // Track for 1sec
                                    self.dsp_output.borrow_mut().wake_up();
                                }
                                _ => {}
                            }
                        }
                    }
                    NoteEvent::NoteOff {
                        timing: _,
                        voice_id: _,
                        channel,
                        note,
                        velocity: _,
                    } => {
                        if let Some(midi_drum_note) = Self::map_midi_note(note, channel) {
                            match midi_drum_note {
                                GeneralMidiDrums::BassDrum | GeneralMidiDrums::AcousticBassDrum => {
                                    self.dsp_kick.borrow_mut().note_off();
                                }
                                GeneralMidiDrums::AcousticSnare
                                | GeneralMidiDrums::ElectricSnare => {
                                    self.dsp_snare.borrow_mut().note_off();
                                }
                                GeneralMidiDrums::SideStick => {
                                    self.dsp_rimshot1.borrow_mut().note_off();
                                    self.dsp_rimshot2.borrow_mut().note_off();
                                }
                                GeneralMidiDrums::HandClap => {
                                    self.dsp_clap1.borrow_mut().note_off();
                                    self.dsp_clap2.borrow_mut().note_off();
                                }
                                GeneralMidiDrums::LowFloorTom | GeneralMidiDrums::HighFloorTom => {
                                    self.dsp_lowtom1.borrow_mut().note_off();
                                    self.dsp_lowtom2.borrow_mut().note_off();
                                    self.dsp_lowtom3.borrow_mut().note_off();
                                    self.dsp_lowtom4.borrow_mut().note_off();
                                }
                                GeneralMidiDrums::LowTom | GeneralMidiDrums::LowMidTom => {
                                    self.dsp_midtom1.borrow_mut().note_off();
                                    self.dsp_midtom2.borrow_mut().note_off();
                                    self.dsp_midtom3.borrow_mut().note_off();
                                    self.dsp_midtom4.borrow_mut().note_off();
                                }
                                GeneralMidiDrums::HighMidTom | GeneralMidiDrums::HighTom => {
                                    self.dsp_hitom1.borrow_mut().note_off();
                                    self.dsp_hitom2.borrow_mut().note_off();
                                    self.dsp_hitom3.borrow_mut().note_off();
                                    self.dsp_hitom4.borrow_mut().note_off();
                                }
                                GeneralMidiDrums::ClosedHihat => {
                                    self.dsp_hhclosed.borrow_mut().note_off();
                                }
                                GeneralMidiDrums::PedalHihat => {}
                                GeneralMidiDrums::OpenHihat => {
                                    self.dsp_hhopen.borrow_mut().note_off();
                                }
                                GeneralMidiDrums::CrashCymbal | GeneralMidiDrums::RideCymbal => {
                                    self.dsp_cymbal1.borrow_mut().note_off();
                                    self.dsp_cymbal2.borrow_mut().note_off();
                                    self.dsp_cymbal3.borrow_mut().note_off();
                                }
                                GeneralMidiDrums::Cowbell | GeneralMidiDrums::Tambourine => {
                                    self.dsp_cowbell1.borrow_mut().note_off();
                                    self.dsp_cowbell2.borrow_mut().note_off();
                                    self.dsp_cowbell3.borrow_mut().note_off();
                                }
                                _ => {}
                            }
                        }
                    }
                    _ => (),
                }
            },
            #[allow(unused_variables)]
            &mut |buffer: &mut Buffer, aux: &mut AuxiliaryBuffers, block_start, block_end| // AUDIO
            {
                //let frames = output[0].len() as i32;
                let frames = block_end - block_start;

                // Attempt to process audio in blocks split by events:
                let output = &mut buffer.slice2ch_range_mut(block_start, block_end);
                for channel in output.iter_mut() {
                    channel.fill(0.0);
                }

                //self.accum_buffer.clear();
                self.accum_buffer.buffer.clear_frames(frames);

                for_each_voice_dsp!(self, &mut |dsp: &mut dyn OneTrickDSPGeneral| {
                    // Compute Voice:
                    if dsp.is_active() {
                        dsp.compute(frames, &[]);
                    } else {
                        dsp.skip_compute();
                    }
                    // Accumulate Voice:
                    dsp.add_to_buffer(&mut self.accum_buffer.buffer);
                });

                // Process Main Output:
                {
                    let mut dsp = self.dsp_output.borrow_mut();
                    if dsp.is_active() {
                        dsp.compute_to(frames, &self.accum_buffer.buffer.as_slice_actually_immutable(), Some(output));
                    } else {
                        dsp.skip_compute();
                    }
                }

                // Handle Aux Output:
                #[cfg(feature = "multi_out") ]
                {
                    // NIH Plug will zero out Aux buffers is it gets "weird data" from host...
                    // Ardour: Uses only the first Aux output..
                    let aux_supported: bool = !aux.outputs.is_empty()
                        && !aux.outputs[0].is_empty()
                        && !aux.outputs[0].as_slice()[0].is_empty();
                    if aux_supported {

                        let mut aux_outs = aux.outputs.iter_mut();

                        for_each_voice_dsp_grouped!(self, &mut |group: &[&mut dyn OneTrickDSPGeneral]| {
                            if let Some(aux_out) = aux_outs.next() {
                                if !aux_out.is_empty() && !aux_out.as_slice()[0].is_empty() {
                                    let mut is_first = true;
                                    for dsp in group {
                                        if dsp.compute_ran() {
                                            let mut slice = aux_out.slice2ch_range_mut(block_start, block_end);
                                            if is_first {
                                                dsp.write_to_slice(&mut slice);
                                                is_first = false;
                                            } else {
                                                dsp.add_to_slice(&mut slice);
                                            }
                                        }
                                    }
                                }
                            }
                        });
                    }
                }
            },
        );

        #[cfg(feature = "profile")]
        {
            let sec: f64 = dsp_timer.elapsed().as_secs_f64().try_into().unwrap();
            let max_sec: f64 = buffer.samples() as f64 / self.sample_rate as f64;
            let ratio_int = (sec / max_sec * 100. * 100000.) as i32; // Put it into %, then 5 decimals of precision
                                                                     // VU-style falling:
            let old_ratio_int = self.dsp_cost.load(Ordering::Relaxed);
            let ratio_int = ratio_int.max(old_ratio_int - 2000);
            self.dsp_cost.store(ratio_int, Ordering::Relaxed);
        }

        ProcessStatus::Normal
    }

    #[cfg(feature = "egui")]
    fn editor(&mut self, _async_executor: AsyncExecutor<Self>) -> Option<Box<dyn Editor>> {
        let params = self.params.clone();
        
        //println!("editor()");

        // let art_svg:Arc<AtomicRefCell<Option<RetainedImage>>> = Arc::new(AtomicRefCell::new(None));
        // let art_svg_loader = art_svg.clone();
        
        // let logo_blocker_svg: Arc<AtomicRefCell<Option<RetainedImage>>> = Arc::new(AtomicRefCell::new(None));
        // let logo_blocker_svg_loader = logo_blocker_svg.clone();
        
        // let grunge1_svg: Arc<AtomicRefCell<Option<RetainedImage>>> = Arc::new(AtomicRefCell::new(None));
        // let grunge1_svg_loader = grunge1_svg.clone();

        // let grunge2_svg: Arc<AtomicRefCell<Option<RetainedImage>>> = Arc::new(AtomicRefCell::new(None));
        // let grunge2_svg_loader = grunge2_svg.clone();

        let palette = Palette::new(15)
            .shades(6)
            .white_level(1.0)
            .black_level(0.1)
            .saturation(0.85)
            .shift(-0.075+0.4)
            .alt_hue_step(1.5)
            .dark_shift(0.1)
            .dark_desaturation(0.5);
        //let palette_alt = palette.alternate(1.5);
        
        let mut knob_style = ParamKnobStyle {
            radius: 35.0,
            indicator_style: ParamKnobIndicatorStyle::Continuous,
            ..Default::default()
        };
        knob_style.knob_color = palette.grey().into();
        knob_style.label_style.bg_color = Color32::from_black_alpha(96);
        knob_style.bg_color = palette.black().into();
        knob_style.indicator_fill_color = palette.grey().into();
        let column_width = knob_style.required_width();

        let mut knob_style_1 = knob_style.clone();
        knob_style_1.indicator_fill_color = palette.color(1).complimentary().into();
        //knob_style_1.knob_color = palette_alt.color(1).complimentary().into();
        let mut knob_style_2 = knob_style.clone();
        knob_style_2.indicator_fill_color = palette.color(0).into();
        //knob_style_2.knob_color = palette_alt.color(0).into();
        let mut knob_style_3 = knob_style.clone();
        knob_style_3.indicator_fill_color = palette.color(1).into();
        //knob_style_3.knob_color = palette_alt.color(1).into();
        let mut knob_style_4 = knob_style.clone();
        knob_style_4.indicator_fill_color = palette.color(2).into();
        //knob_style_4.knob_color = palette_alt.color(2).into();
        let mut knob_style_5 = knob_style.clone();
        knob_style_5.indicator_fill_color = palette.color(3).into();
        //knob_style_5.knob_color = palette_alt.color(3).into();

        let mut slider_style = ParamSliderStyle {
            length: 70.0,
            ..Default::default()
        };
        slider_style.knob_color = knob_style.knob_color;
        slider_style.label_width = 70.0;
        slider_style.label_style.bg_color = Color32::from_black_alpha(96);
        slider_style.bg_color = palette.black().into();
        slider_style.fill_color = Color32::LIGHT_GREEN;

        let mut slider_style_1 = slider_style.clone();
        slider_style_1.fill_color = knob_style_1.indicator_fill_color;
        let mut slider_style_2 = slider_style.clone();
        slider_style_2.fill_color = knob_style_2.indicator_fill_color;
        let mut slider_style_3 = slider_style.clone();
        slider_style_3.fill_color = knob_style_3.indicator_fill_color;
        let mut slider_style_4 = slider_style.clone();
        slider_style_4.fill_color = knob_style_4.indicator_fill_color;
        #[allow(clippy::redundant_clone)]
        let mut slider_style_5 = slider_style.clone();
        slider_style_5.fill_color = knob_style_5.indicator_fill_color;

        let toggle_length = 45.0;

        let mut heading_style = LabelStyle::default_heading();
        heading_style.width = Some(column_width);
        heading_style.color = palette.black().into();
        heading_style.bg_color = palette.white().brightness(0.1).into(); //Color32::from_gray(128);
        let mut heading_style_1 = heading_style.clone();
        heading_style_1.bg_color = palette.color(1).complimentary().into(); // Overridden later!
        let mut heading_style_2 = heading_style.clone();
        heading_style_2.bg_color = palette.color(0).into(); // Overridden later!
        let mut heading_style_3 = heading_style.clone();
        heading_style_3.bg_color = palette.color(1).into(); // Overridden later!
        let mut heading_style_4 = heading_style.clone();
        heading_style_4.bg_color = palette.color(2).into(); // Overridden later!
        let mut heading_style_5 = heading_style.clone();
        heading_style_5.bg_color = palette.color(4).into(); // Overridden later!

        heading_style.color = palette.white().into();

        let show_credits = Arc::new(AtomicBool::new(false));
        //let anim_id = Id::new("local_anim"); // Can use in each context once
        let credits_anim_id = Id::new("credits_anim");
        let kick_anim_id = Id::new("kick_anim");
        let snare_anim_id = Id::new("snare_anim");
        let rimshot_anim_id = Id::new("rimshot_anim");
        let clap_anim_id = Id::new("clap_anim");
        let lowtom_anim_id = Id::new("lowtom_anim");
        let midtom_anim_id = Id::new("midtom_anim");
        let hitom_anim_id = Id::new("hitom_anim");
        let hhclosed_anim_id = Id::new("hhclosed_anim");
        let hhopen_anim_id = Id::new("hhopen_anim");
        let cymbal_anim_id = Id::new("cymbal_anim");
        let cowbell_anim_id = Id::new("cowbell_anim");

        #[cfg(feature = "profile")]
        let dsp_cost = self.dsp_cost.clone();
        
        let kick_indicator = self.kick_indicator.clone();
        let snare_indicator = self.snare_indicator.clone();
        let rimshot_indicator = self.rimshot_indicator.clone();
        let clap_indicator = self.clap_indicator.clone();
        let lowtom_indicator = self.lowtom_indicator.clone();
        let midtom_indicator = self.midtom_indicator.clone();
        let hitom_indicator = self.hitom_indicator.clone();
        let hhclosed_indicator = self.hhclosed_indicator.clone();
        let hhopen_indicator = self.hhopen_indicator.clone();
        let cymbal_indicator = self.cymbal_indicator.clone();
        let cowbell_indicator = self.cowbell_indicator.clone();

        create_egui_editor(
            self.params.editor_state.clone(),
            (),
            move |ctx, _| {
                // DPI:
                // egui_baseview needs to be updated to support DPI.
                // We can adjust egui's rendering, but not egui_baseview's resolution.
                //ctx.set_pixels_per_point(2.0);
                
                //println!("Editor::Open()");
                Self::setup_egui_fonts(ctx);

                // *art_svg_loader.borrow_mut() = RetainedImage::from_svg_bytes(
                //     "CryptidArt1.svg",
                //     include_bytes!("assets/svgs/CryptidArt1.svg")).ok();

                // *logo_blocker_svg_loader.borrow_mut() = RetainedImage::from_svg_bytes(
                //     "LogoBlockerSplatter.svg",
                //     include_bytes!("../onetrick/src/assets/svgs/LogoBlockerSplatter.svg")).ok();

                // *grunge1_svg_loader.borrow_mut() = RetainedImage::from_svg_bytes(
                //     "Grunge1.svg",
                //     include_bytes!("../onetrick/src/assets/svgs/Grunge1.svg")).ok();

                // *grunge2_svg_loader.borrow_mut() = RetainedImage::from_svg_bytes(
                //     "Grunge2.svg",
                //     include_bytes!("../onetrick/src/assets/svgs/Grunge2.svg")).ok();


            },
            move |ctx, setter, _state| {
                //update()
                //let window_rect = ctx.available_rect();

                #[cfg(feature = "profile")]
                let dsp_cost = dsp_cost.load(Ordering::Relaxed) as f32 * 0.00001; // Scale for 5 decimals of precision

                let label_brightness = 0.85;
                let label_darken_amount = 0.5;
                let indicator_raw = kick_indicator.swap(false, Ordering::Relaxed);
                let kick_indicator_brightness = label_brightness-ctx.animate_bool_with_time(
                    kick_anim_id,
                    indicator_raw,
                    if indicator_raw { 0.0 } else { 0.300 },
                ) * label_darken_amount;
                let indicator_raw = snare_indicator.swap(false, Ordering::Relaxed);
                let snare_indicator_brightness = label_brightness-ctx.animate_bool_with_time(
                    snare_anim_id,
                    indicator_raw,
                    if indicator_raw { 0.0 } else { 0.300 },
                ) * label_darken_amount;
                let indicator_raw = rimshot_indicator.swap(false, Ordering::Relaxed);
                let rimshot_indicator_brightness = label_brightness-ctx.animate_bool_with_time(
                    rimshot_anim_id,
                    indicator_raw,
                    if indicator_raw { 0.0 } else { 0.300 },
                ) * label_darken_amount;
                let indicator_raw = clap_indicator.swap(false, Ordering::Relaxed);
                let clap_indicator_brightness = label_brightness-ctx.animate_bool_with_time(
                    clap_anim_id,
                    indicator_raw,
                    if indicator_raw { 0.0 } else { 0.300 },
                ) * label_darken_amount;
                let indicator_raw = lowtom_indicator.swap(false, Ordering::Relaxed);
                let lowtom_indicator_brightness = label_brightness-ctx.animate_bool_with_time(
                    lowtom_anim_id,
                    indicator_raw,
                    if indicator_raw { 0.0 } else { 0.300 },
                ) * label_darken_amount;
                let indicator_raw = midtom_indicator.swap(false, Ordering::Relaxed);
                let midtom_indicator_brightness = label_brightness-ctx.animate_bool_with_time(
                    midtom_anim_id,
                    indicator_raw,
                    if indicator_raw { 0.0 } else { 0.300 },
                ) * label_darken_amount;
                let indicator_raw = hitom_indicator.swap(false, Ordering::Relaxed);
                let hitom_indicator_brightness = label_brightness-ctx.animate_bool_with_time(
                    hitom_anim_id,
                    indicator_raw,
                    if indicator_raw { 0.0 } else { 0.300 },
                ) * label_darken_amount;
                let indicator_raw = hhclosed_indicator.swap(false, Ordering::Relaxed);
                let hhclosed_indicator_brightness = label_brightness-ctx.animate_bool_with_time(
                    hhclosed_anim_id,
                    indicator_raw,
                    if indicator_raw { 0.0 } else { 0.300 },
                ) * label_darken_amount;
                let indicator_raw = hhopen_indicator.swap(false, Ordering::Relaxed);
                let hhopen_indicator_brightness = label_brightness-ctx.animate_bool_with_time(
                    hhopen_anim_id,
                    indicator_raw,
                    if indicator_raw { 0.0 } else { 0.300 },
                ) * label_darken_amount;
                let indicator_raw = cymbal_indicator.swap(false, Ordering::Relaxed);
                let cymbal_indicator_brightness = label_brightness-ctx.animate_bool_with_time(
                    cymbal_anim_id,
                    indicator_raw,
                    if indicator_raw { 0.0 } else { 0.300 },
                ) * label_darken_amount;
                let indicator_raw = cowbell_indicator.swap(false, Ordering::Relaxed);
                let cowbell_indicator_brightness = label_brightness-ctx.animate_bool_with_time(
                    cowbell_anim_id,
                    indicator_raw,
                    if indicator_raw { 0.0 } else { 0.300 },
                ) * label_darken_amount;

                let show_credits_amount =
                    ctx.animate_bool_with_time(credits_anim_id, show_credits.load(Ordering::Relaxed), 0.300);

                if show_credits_amount > 0.0 {
                    Area::new("credits_area".into())
                        .fixed_pos(Pos2::new(0.0, 0.0))
                        .order(Order::Foreground)
                        .show(ctx, |ui| {
                            Frame::none()
                                .outer_margin(Margin::same(0.0))
                                .inner_margin(Margin::same(10.0))
                                .fill(Color32::from_black_alpha(
                                    (240.0 * show_credits_amount) as u8,
                                ))
                                //.rounding(Rounding::same(20.0))
                                .show(ui, |ui| {
                                    let available_size =
                                        Vec2::new(ui.available_width(), ui.available_height());
                                    //let animated_offset = Vec2::new(0.0, available_size.y * (1.0-show_credits_amount));
                                    ui.painter().text(
                                        (available_size * 0.5).to_pos2(),
                                        Align2::CENTER_CENTER,
                                        format!(
                                            include_str!("CREDITS"),
                                            VERSION = env!("CARGO_PKG_VERSION")
                                        ),
                                        FontId::new(24.0, FontFamily::Name("Title".into())),
                                        Color32::from_white_alpha(
                                            (255.0 * show_credits_amount) as u8,
                                        ),
                                    );
                                    /*
                                    let galley = ui.painter().layout_no_wrap(
                                        include_str!("CREDITS").to_string(),
                                        FontId::new(20.0, FontFamily::Name("Title".into())),
                                        Color32::WHITE);
                                    let rect = Align2::CENTER_CENTER.anchor_rect(Rect::from_min_size((available_size * 0.5).to_pos2(), galley.size()));
                                    ui.painter().galley(rect.min, galley);
                                    */

                                    if ui
                                        .allocate_response(available_size, Sense::click())
                                        .clicked()
                                    {
                                        show_credits.store(false, Ordering::Relaxed);
                                    }
                                });
                        });
                }

                let top_bg_color: Color32 = palette.white().brightness(0.05).into();
                /*
                    //palette.color(2).shade(2).into()
                    //palette.black().into()
                    palette.white().brightness(0.05).into(),
                */
                TopBottomPanel::top("top")
                    .frame(
                        Frame::none()
                            .outer_margin(Margin::same(0.0))
                            .inner_margin(Margin::same(10.0))
                            .fill(
                                top_bg_color
                            ), //.rounding(Rounding::same(20.0))
                               //.rounding(Rounding{ne: 20.0, nw: 20.0, ..Default::default()})
                    )
                    .show(ctx, |ui| {
                        // let art_svg = art_svg.borrow();
                        // ui.painter().svg(ctx, &art_svg.as_ref().unwrap(), Pos2::new(210.0, 80.0), 215.0, palette.white().brightness(1.0).into());
                        
                        // let logo_blocker_svg = logo_blocker_svg.borrow();
                        // ui.painter().svg(ctx, &logo_blocker_svg.as_ref().unwrap(), Pos2::new(184.0, 62.0), 105.0, top_bg_color);
                        
                        Area::new("logo_area".into()) // Prevents layout being affected by Logo
                        .fixed_pos(Pos2::new(0.0, 0.0))
                        .order(Order::Foreground)
                        .show(ctx, |ui| {
                            let response = ui.allocate_rect(
                                Rect::from_min_size(Pos2::new(0.0, 0.0),
                                Vec2::new(380.0, 120.0)),
                                Sense::click());
                            if response.clicked() {
                                show_credits.store(true, Ordering::Relaxed);
                            }
                            ui.painter().one_trick_logo(
                                "CRYPTID",
                                Pos2::new(55.0, 60.0),
                                65.0,
                                palette.color(1).complimentary().into(),
                                if response.hovered() {palette.white().into()} else {palette.color(1).into()},
                            );
                            ui.painter().fade_80s_uniform(Rect::from_min_size(Pos2::new(120.0, 60.0), Vec2::new(225.0, 20.0)), 1.0, top_bg_color);
                        });
                        ui.painter().fade_80s(Rect::from_min_size(Pos2::new(0.0, 117.25), Vec2::new(1200.0, 50.0)), 10, palette.color(1).complimentary().brightness(0.6).into());
                        ui.set_width(ui.available_width());
                        ui.set_height(100.0);
                        //ui.with_layout(Layout::right_to_left(Align::TOP), |ui| {
                        let panel_shadow = Shadow{
                            offset:Vec2::new(0.0, 0.0),
                            spread: 0.0, 
                            blur: 15.0,
                            color: palette.black().alpha(0.6).into()
                        };
                        ui.horizontal(|ui| {
                            ui.add_space(340.0+75.0);
                            Frame::none() // Global Reverb
                                .outer_margin(Margin::same(0.0))
                                .inner_margin(Margin::same(10.0))
                                .fill(palette.white().brightness(0.35).into())
                                .rounding(Rounding::same(20.0))
                                .shadow(panel_shadow)
                                .show(ui, |ui| {
                                    ui.vertical(|ui| {
                                        ui.horizontal(|ui| {
                                            ui.set_width(75.0 * 5.0);
                                            ui.centered_and_justified(|ui| {
                                                ui.add(
                                                    Label::new("Reverb").with_style(&heading_style),
                                                );
                                            });
                                        });
                                        ui.add_space(5.0);
                                        ui.horizontal(|ui| {
                                            /*
                                            ui.add(
                                                ParamKnob::for_param(
                                                    params.param_float("Global Reverb Time"),
                                                    setter,
                                                )
                                                .with_label("Time")
                                                .with_style(&knob_style),
                                            );
                                            */
                                            ui.add(
                                                ParamKnob::for_param(
                                                    params.param_float("Global Reverb Size"),
                                                    setter,
                                                )
                                                .with_label("Size")
                                                .with_style(&knob_style),
                                            );
                                            ui.add(
                                                ParamKnob::for_param(
                                                    params.param_float("Global Reverb Diffusion"),
                                                    setter,
                                                )
                                                .with_label("Diff")
                                                .with_style(&knob_style),
                                            );
                                            ui.add(
                                                ParamKnob::for_param(
                                                    params.param_float("Global Reverb Damp"),
                                                    setter,
                                                )
                                                .with_label("Damp")
                                                .with_style(&knob_style),
                                            );
                                            ui.add(
                                                ParamKnob::for_param(
                                                    params.param_float("Global Reverb Gate"),
                                                    setter,
                                                )
                                                .with_label("Gate")
                                                .with_style(&knob_style),
                                            );
                                            ui.add(
                                                ParamKnob::for_param(
                                                    params.param_float("Global Reverb"),
                                                    setter,
                                                )
                                                .with_label("Amount")
                                                .with_style(&knob_style_3),
                                            );
                                        });
                                    });
                                });
                            Frame::none() // Globals
                                .outer_margin(Margin::same(0.0))
                                .inner_margin(Margin::same(10.0))
                                .fill(palette.white().brightness(0.35).into())
                                .rounding(Rounding::same(20.0))
                                .shadow(panel_shadow)
                                .show(ui, |ui| {
                                    ui.vertical(|ui| {
                                        ui.horizontal(|ui| {
                                            ui.set_width(75.0 * 3.0);
                                            ui.centered_and_justified(|ui| {
                                                ui.add(
                                                    Label::new("Global").with_style(&heading_style),
                                                );
                                            });
                                        });
                                        ui.add_space(5.0);
                                        ui.horizontal(|ui| {
                                            ui.add(
                                                ParamKnob::for_param(
                                                    params.param_float("Global Sensitivity"),
                                                    setter,
                                                )
                                                .with_label("Velocity")
                                                .with_style(&knob_style),
                                            );
                                            ui.add(
                                                ParamKnob::for_param(
                                                    params.param_float("Global Saturation"),
                                                    setter,
                                                )
                                                .with_label("Saturation")
                                                .with_style(&knob_style),
                                            );
                                            /*
                                            ui.add(
                                                ParamKnob::for_param(
                                                    params.param_float("Global Alt Transpose"),
                                                    setter,
                                                )
                                                .with_label("Alt Tune")
                                                .with_style(&knob_style)
                                                //.with_indicator_from_center(),
                                            );
                                            */
                                            ui.add(
                                                ParamKnob::for_param(
                                                    params.param_float("Global Volume"),
                                                    setter,
                                                )
                                                .with_label("Volume")
                                                .with_style(&knob_style_3),
                                            );
                                        });
                                    });
                                });

                        });

                        #[cfg(feature = "profile")]
                        {
                            let mut rect =
                                Rect::from_center_size(Pos2::new(120., 120.), Vec2::new(200., 20.));
                            ui.painter()
                                .rect_filled(rect, Rounding::same(3.0), Color32::DARK_GRAY);
                            let ratio_of_budget = dsp_cost * 0.01 / 0.05; // Set budge to 5%...
                            rect.set_right(rect.left() + rect.width() * ratio_of_budget.min(1.0));
                            ui.painter().rect_filled(
                                rect,
                                Rounding::same(3.0),
                                if ratio_of_budget >= 1.0 {
                                    Color32::RED
                                } else {
                                    Color32::GRAY
                                },
                            );

                            ui.painter().text(
                                Pos2::new(20.0, 120.0),
                                Align2::LEFT_CENTER,
                                format!("{:.1}%", ratio_of_budget * 100.),
                                FontId::new(20.0, FontFamily::Name("Title".into())),
                                Color32::WHITE,
                            );
                        }
                    });
                /*
                TopBottomPanel::bottom("bottom")
                    .frame(
                        Frame::none()
                            .outer_margin(Margin::same(0.0))
                            .inner_margin(Margin::same(10.0))
                            .fill(Color32::DARK_RED)
                            //.rounding(Rounding{se: 20.0, sw: 20.0, ..Default::default()})
                    )
                    .show(ctx, |ui|
                {
                    ui.set_width(ui.available_width());
                    ui.set_height(100.0);
                });
                */
                CentralPanel::default()
                    .frame(
                        Frame::none()
                            .outer_margin(Margin::same(0.0))
                            .inner_margin(Margin::same(10.0))
                            .fill(palette.white().brightness(0.30).into()), //.rounding(Rounding{se: 20.0, sw: 20.0, ..Default::default()})
                    )
                    .show(ctx, |ui| {
                        ui.visuals_mut().widgets.noninteractive.bg_stroke =
                            Stroke::new(2.0, palette.grey().alpha(0.666).to_color32());
                        //ctx.set_debug_on_hover(false);
                        // let grunge1_svg = grunge1_svg.borrow();
                        // ui.painter().svg(ctx,
                        //     &grunge1_svg.as_ref().unwrap(),
                        //     Pos2::new(250.0, window_rect.height() - 170.0),
                        //     300.0,
                        //     palette.black().alpha(0.33).into());
                        // let grunge2_svg = grunge2_svg.borrow();
                        // ui.painter().svg(ctx,
                        //     &grunge2_svg.as_ref().unwrap(),
                        //     Pos2::new(760.0, window_rect.height() - 50.0),
                        //     300.0,
                        //     palette.black().alpha(0.33).into());
                        ui.set_width(ui.available_width());
                        ui.horizontal(|ui| {
                            let bottom_pos = ui.next_widget_position().y + 400.0 + 95. + 17.;
                            ui.spacing_mut().item_spacing = vec2(10.0, 10.0);
                            ui.add_space(20.0); // Centering isn't working, so fake it...
                            ui.vertical(|ui| {
                                ui.set_width(column_width);
                                ui.vertical_centered(|ui| {
                                    ui.add(
                                        Label::new("Kick")
                                            .with_style(&heading_style_1)
                                            .with_bg_color(
                                                palette
                                                    .color(1).complimentary()
                                                    .brightness(kick_indicator_brightness)
                                                    .into(),
                                            ),
                                    );
                                });
                                ui.add(
                                    ParamKnob::for_param(params.param_float("Kick Gain"), setter)
                                        .with_label("Gain")
                                        .with_style(&knob_style_1),
                                );
                                ui.add(
                                    ParamSlider::for_param(params.param_float("Kick Pan"), setter)
                                        .with_label("Pan")
                                        .with_style(&slider_style_1)
                                        .with_indicator_from_center()
                                        .with_orientation(ParamSliderOrientation::Horizontal),
                                );
                                ui.add(
                                    ParamKnob::for_param(params.param_float("Kick Reverb"), setter)
                                        .with_label("Reverb")
                                        .with_style(&knob_style_1),
                                );
                                ui.add(
                                    ParamKnob::for_param(
                                        params.param_float("Kick Transpose"),
                                        setter,
                                    )
                                    .with_label("Tune")
                                    .with_style(&knob_style_1),
                                );
                                ui.add(
                                    ParamKnob::for_param(
                                        params.param_float("Kick Thumpy"),
                                        setter,
                                    )
                                    .with_label("Thumpy")
                                    .with_style(&knob_style_1),
                                );
                                ui.add(
                                    ParamKnob::for_param(params.param_float("Kick Decay"), setter)
                                        .with_label("Decay")
                                        .with_style(&knob_style_1),
                                );
                            });
                            ui.separator();
                            ui.vertical(|ui| {
                                ui.set_width(column_width);
                                ui.vertical_centered(|ui| {
                                    ui.add(
                                        Label::new("Snare")
                                            .with_style(&heading_style_1)
                                            .with_bg_color(
                                                palette
                                                    .color(1).complimentary()
                                                    .brightness(snare_indicator_brightness)
                                                    .into(),
                                            ),
                                    );
                                });
                                ui.add(
                                    ParamKnob::for_param(params.param_float("Snare Gain"), setter)
                                        .with_label("Gain")
                                        .with_style(&knob_style_1),
                                );
                                ui.add(
                                    ParamSlider::for_param(params.param_float("Snare Pan"), setter)
                                        .with_label("Pan")
                                        .with_style(&slider_style_1)
                                        .with_indicator_from_center()
                                        .with_orientation(ParamSliderOrientation::Horizontal),
                                );
                                ui.add(
                                    ParamKnob::for_param(
                                        params.param_float("Snare Reverb"),
                                        setter,
                                    )
                                    .with_label("Reverb")
                                    .with_style(&knob_style_1),
                                );
                                ui.add(
                                    ParamKnob::for_param(
                                        params.param_float("Snare Transpose"),
                                        setter,
                                    )
                                    .with_label("Tune")
                                    .with_style(&knob_style_1),
                                );
                                ui.add(
                                    ParamKnob::for_param(
                                        params.param_float("Snare Snappy"),
                                        setter,
                                    )
                                    .with_label("Snappy")
                                    .with_style(&knob_style_1),
                                );
                                ui.add(
                                    ParamKnob::for_param(
                                        params.param_float("Snare Tone"),
                                        setter,
                                    )
                                    .with_label("Filter")
                                    .with_style(&knob_style_1),
                                );
                            });
                            ui.separator();
                            ui.vertical(|ui| {
                                ui.set_width(column_width);
                                ui.vertical_centered(|ui| {
                                    let title =
                                        if let Some(param) = params.param_int("Rimshot Voice") {
                                            param.to_string()
                                        } else {
                                            "???".to_string()
                                        };
                                    ui.add(
                                        Label::new(title)
                                            .with_style(&heading_style_2)
                                            .with_bg_color(
                                                palette
                                                    .color(0)
                                                    .brightness(rimshot_indicator_brightness)
                                                    .into(),
                                            ),
                                    );
                                });
                                ui.add(
                                    ParamKnob::for_param(
                                        params.param_float("Rimshot Gain"),
                                        setter,
                                    )
                                    .with_label("Gain")
                                    .with_style(&knob_style_2),
                                );
                                ui.add(
                                    ParamSlider::for_param(
                                        params.param_float("Rimshot Pan"),
                                        setter,
                                    )
                                    .with_label("Pan")
                                    .with_style(&slider_style_2)
                                    .with_indicator_from_center()
                                    .with_orientation(ParamSliderOrientation::Horizontal),
                                );
                                ui.add(
                                    ParamKnob::for_param(
                                        params.param_float("Rimshot Reverb"),
                                        setter,
                                    )
                                    .with_label("Reverb")
                                    .with_style(&knob_style_2),
                                );
                                ui.add_space(bottom_pos - ui.next_widget_position().y); //Push to bottom
                                ui.add(
                                    ParamSlider::for_param(
                                        params.param_int("Rimshot Voice"),
                                        setter,
                                    )
                                    .with_label("Voice")
                                    .with_style(&slider_style_2)
                                    .with_tick_count(2)
                                    .without_alt_tick_length()
                                    .without_fill()
                                    //.with_label_only_value()
                                    .with_length(toggle_length),
                                );
                            });
                            ui.separator();
                            ui.vertical(|ui| {
                                ui.set_width(column_width);
                                let (title, is_maraca) =
                                    if let Some(param) = params.param_int("Clap Voice") {
                                        (param.to_string(), param.value()==1)
                                    } else {
                                        ("???".to_string(), false)
                                    };
                                ui.vertical_centered(|ui| {
                                    ui.add(
                                        Label::new(title)
                                            .with_style(&heading_style_2)
                                            .with_bg_color(
                                                palette
                                                    .color(0)
                                                    .brightness(clap_indicator_brightness)
                                                    .into(),
                                            ),
                                    );
                                });
                                ui.add(
                                    ParamKnob::for_param(params.param_float("Clap Gain"), setter)
                                        .with_label("Gain")
                                        .with_style(&knob_style_2),
                                );
                                ui.add(
                                    ParamSlider::for_param(params.param_float("Clap Pan"), setter)
                                        .with_label("Pan")
                                        .with_style(&slider_style_2)
                                        .with_indicator_from_center()
                                        .with_orientation(ParamSliderOrientation::Horizontal),
                                );
                                ui.add(
                                    ParamKnob::for_param(params.param_float("Clap Reverb"), setter)
                                        .with_label("Reverb")
                                        .with_style(&knob_style_2),
                                );
                                
                                if is_maraca {
                                    ui.add(
                                        ParamKnob::for_param(
                                            params.param_float("Clap Attack"),
                                            setter,
                                        )
                                        .with_label("Attack")
                                        .with_style(&knob_style_2),
                                    );
                                } else {
                                    ui.add(
                                        ParamKnob::for_param(
                                            params.param_float("Clap Transpose"),
                                            setter,
                                        )
                                        .with_label("Tune")
                                        .with_style(&knob_style_2),
                                    );
                                }

                                
                                ui.add(
                                    ParamKnob::for_param(
                                        params.param_float("Clap Tone"),
                                        setter,
                                    )
                                    .with_label("Filter")
                                    .with_style(&knob_style_2),
                                );
                                ui.add_space(bottom_pos - ui.next_widget_position().y); //Push to bottom
                                ui.add(
                                    ParamSlider::for_param(params.param_int("Clap Voice"), setter)
                                        .with_label("Voice")
                                        .with_style(&slider_style_2)
                                        .with_tick_count(2)
                                        .without_alt_tick_length()
                                        .without_fill()
                                        //.with_label_only_value()
                                        .with_length(toggle_length),
                                );
                            });
                            ui.separator();
                            ui.vertical(|ui| {
                                ui.set_width(column_width);
                                ui.vertical_centered(|ui| {
                                    let title =
                                        if let Some(param) = params.param_int("Low Tom Voice") {
                                            param.to_string()
                                        } else {
                                            "???".to_string()
                                        };
                                    ui.add(
                                        Label::new(title)
                                            .with_style(&heading_style_3)
                                            .with_bg_color(
                                                palette
                                                    .color(1)
                                                    .brightness(lowtom_indicator_brightness)
                                                    .into(),
                                            ),
                                    );
                                });
                                ui.add(
                                    ParamKnob::for_param(
                                        params.param_float("Low Tom Gain"),
                                        setter,
                                    )
                                    .with_label("Gain")
                                    .with_style(&knob_style_3),
                                );
                                ui.add(
                                    ParamSlider::for_param(
                                        params.param_float("Low Tom Pan"),
                                        setter,
                                    )
                                    .with_label("Pan")
                                    .with_style(&slider_style_3)
                                    .with_indicator_from_center()
                                    .with_orientation(ParamSliderOrientation::Horizontal),
                                );
                                ui.add(
                                    ParamKnob::for_param(
                                        params.param_float("Low Tom Reverb"),
                                        setter,
                                    )
                                    .with_label("Reverb")
                                    .with_style(&knob_style_3),
                                );
                                ui.add(
                                    ParamKnob::for_param(
                                        params.param_float("Low Tom Transpose"),
                                        setter,
                                    )
                                    .with_label("Tune")
                                    .with_style(&knob_style_3),
                                );
                                ui.add(
                                    ParamKnob::for_param(
                                        params.param_float("Low Tom Alt Transpose"),
                                        setter,
                                    )
                                    .with_label("Alt Tune")
                                    .with_style(&knob_style)
                                    //.with_indicator_from_center(),
                                );
                                ui.add_space(bottom_pos - ui.next_widget_position().y); //Push to bottom
                                ui.add(
                                    ParamSlider::for_param(
                                        params.param_int("Low Tom Voice"),
                                        setter,
                                    )
                                    .with_label("Voice")
                                    .with_style(&slider_style_3)
                                    .with_tick_count(4)
                                    .without_alt_tick_length()
                                    .without_fill()
                                    //.with_label_only_value()
                                    .with_length(toggle_length),
                                );
                            });
                            ui.separator();
                            ui.vertical(|ui| {
                                ui.set_width(column_width);
                                ui.vertical_centered(|ui| {
                                    let title =
                                        if let Some(param) = params.param_int("Mid Tom Voice") {
                                            param.to_string()
                                        } else {
                                            "???".to_string()
                                        };
                                    ui.add(
                                        Label::new(title)
                                            .with_style(&heading_style_3)
                                            .with_bg_color(
                                                palette
                                                    .color(1)
                                                    .brightness(midtom_indicator_brightness)
                                                    .into(),
                                            ),
                                    );
                                });
                                ui.add(
                                    ParamKnob::for_param(
                                        params.param_float("Mid Tom Gain"),
                                        setter,
                                    )
                                    .with_label("Gain")
                                    .with_style(&knob_style_3),
                                );
                                ui.add(
                                    ParamSlider::for_param(
                                        params.param_float("Mid Tom Pan"),
                                        setter,
                                    )
                                    .with_label("Pan")
                                    .with_style(&slider_style_3)
                                    .with_indicator_from_center()
                                    .with_orientation(ParamSliderOrientation::Horizontal),
                                );
                                ui.add(
                                    ParamKnob::for_param(
                                        params.param_float("Mid Tom Reverb"),
                                        setter,
                                    )
                                    .with_label("Reverb")
                                    .with_style(&knob_style_3),
                                );
                                ui.add(
                                    ParamKnob::for_param(
                                        params.param_float("Mid Tom Transpose"),
                                        setter,
                                    )
                                    .with_label("Tune")
                                    .with_style(&knob_style_3),
                                );
                                ui.add(
                                    ParamKnob::for_param(
                                        params.param_float("Mid Tom Alt Transpose"),
                                        setter,
                                    )
                                    .with_label("Alt Tune")
                                    .with_style(&knob_style)
                                    //.with_indicator_from_center(),
                                );
                                ui.add_space(bottom_pos - ui.next_widget_position().y); //Push to bottom
                                ui.add(
                                    ParamSlider::for_param(
                                        params.param_int("Mid Tom Voice"),
                                        setter,
                                    )
                                    .with_label("Voice")
                                    .with_style(&slider_style_3)
                                    .with_tick_count(4)
                                    .without_alt_tick_length()
                                    .without_fill()
                                    //.with_label_only_value()
                                    .with_length(toggle_length),
                                );
                            });
                            ui.separator();
                            ui.vertical(|ui| {
                                ui.set_width(column_width);
                                ui.vertical_centered(|ui| {
                                    let title =
                                        if let Some(param) = params.param_int("High Tom Voice") {
                                            param.to_string()
                                        } else {
                                            "???".to_string()
                                        };
                                    ui.add(
                                        Label::new(title)
                                            .with_style(&heading_style_3)
                                            .with_bg_color(
                                                palette
                                                    .color(1)
                                                    .brightness(hitom_indicator_brightness)
                                                    .into(),
                                            ),
                                    );
                                });
                                ui.add(
                                    ParamKnob::for_param(
                                        params.param_float("High Tom Gain"),
                                        setter,
                                    )
                                    .with_label("Gain")
                                    .with_style(&knob_style_3),
                                );
                                ui.add(
                                    ParamSlider::for_param(
                                        params.param_float("High Tom Pan"),
                                        setter,
                                    )
                                    .with_label("Pan")
                                    .with_style(&slider_style_3)
                                    .with_indicator_from_center()
                                    .with_orientation(ParamSliderOrientation::Horizontal),
                                );
                                ui.add(
                                    ParamKnob::for_param(
                                        params.param_float("High Tom Reverb"),
                                        setter,
                                    )
                                    .with_label("Reverb")
                                    .with_style(&knob_style_3),
                                );
                                ui.add(
                                    ParamKnob::for_param(
                                        params.param_float("High Tom Transpose"),
                                        setter,
                                    )
                                    .with_label("Tune")
                                    .with_style(&knob_style_3),
                                );
                                ui.add(
                                    ParamKnob::for_param(
                                        params.param_float("High Tom Alt Transpose"),
                                        setter,
                                    )
                                    .with_label("Alt Tune")
                                    .with_style(&knob_style)
                                    //.with_indicator_from_center(),
                                );
                                ui.add_space(bottom_pos - ui.next_widget_position().y); //Push to bottom
                                ui.add(
                                    ParamSlider::for_param(
                                        params.param_int("High Tom Voice"),
                                        setter,
                                    )
                                    .with_label("Voice")
                                    .with_style(&slider_style_3)
                                    .with_tick_count(4)
                                    .without_alt_tick_length()
                                    .without_fill()
                                    //.with_label_only_value()
                                    .with_length(toggle_length),
                                );
                            });
                            //});
                            //ui.add_space(25.0);
                            //ui.horizontal(|ui| {
                            ui.separator();
                            ui.vertical(|ui| {
                                ui.set_width(column_width);
                                ui.vertical_centered(|ui| {
                                    ui.add(
                                        Label::new("Hihat CL")
                                            .with_style(&heading_style_4)
                                            .with_bg_color(
                                                palette
                                                    .color(2)
                                                    .brightness(hhclosed_indicator_brightness)
                                                    .into(),
                                            ),
                                    );
                                });
                                ui.add(
                                    ParamKnob::for_param(
                                        params.param_float("Hihat Closed Gain"),
                                        setter,
                                    )
                                    .with_label("Gain")
                                    .with_style(&knob_style_4),
                                );
                                ui.add(
                                    ParamSlider::for_param(
                                        params.param_float("Hihat Closed Pan"),
                                        setter,
                                    )
                                    .with_label("Pan")
                                    .with_style(&slider_style_4)
                                    .with_indicator_from_center()
                                    .with_orientation(ParamSliderOrientation::Horizontal),
                                );
                                ui.add(
                                    ParamKnob::for_param(
                                        params.param_float("Hihat Closed Reverb"),
                                        setter,
                                    )
                                    .with_label("Reverb")
                                    .with_style(&knob_style_4),
                                );
                                ui.add(
                                    ParamKnob::for_param(
                                        params.param_float("Hihat Closed Tone"),
                                        setter,
                                    )
                                    .with_label("Filter")
                                    .with_style(&knob_style_4),
                                );
                                ui.add(
                                    ParamKnob::for_param(
                                        params.param_float("Hihat Closed Decay"),
                                        setter,
                                    )
                                    .with_label("Decay")
                                    .with_style(&knob_style_4),
                                );
                            });
                            ui.separator();
                            ui.vertical(|ui| {
                                ui.set_width(column_width);
                                ui.vertical_centered(|ui| {
                                    ui.add(
                                        Label::new("Hihat OP")
                                            .with_style(&heading_style_4)
                                            .with_bg_color(
                                                palette
                                                    .color(2)
                                                    .brightness(hhopen_indicator_brightness)
                                                    .into(),
                                            ),
                                    );
                                });
                                ui.add(
                                    ParamKnob::for_param(
                                        params.param_float("Hihat Open Gain"),
                                        setter,
                                    )
                                    .with_label("Gain")
                                    .with_style(&knob_style_4),
                                );
                                ui.add(
                                    ParamSlider::for_param(
                                        params.param_float("Hihat Open Pan"),
                                        setter,
                                    )
                                    .with_label("Pan")
                                    .with_style(&slider_style_4)
                                    .with_indicator_from_center()
                                    .with_orientation(ParamSliderOrientation::Horizontal),
                                );
                                ui.add(
                                    ParamKnob::for_param(
                                        params.param_float("Hihat Open Reverb"),
                                        setter,
                                    )
                                    .with_label("Reverb")
                                    .with_style(&knob_style_4),
                                );
                                ui.add(
                                    ParamKnob::for_param(
                                        params.param_float("Hihat Open Tone"),
                                        setter,
                                    )
                                    .with_label("Filter")
                                    .with_style(&knob_style_4),
                                );
                                ui.add(
                                    ParamKnob::for_param(
                                        params.param_float("Hihat Open Decay"),
                                        setter,
                                    )
                                    .with_label("Decay")
                                    .with_style(&knob_style_4),
                                );
                            });
                            ui.separator();
                            ui.vertical(|ui| {
                                ui.set_width(column_width);
                                ui.vertical_centered(|ui| {
                                    let title =
                                        if let Some(param) = params.param_int("Cymbal Voice") {
                                            param.to_string()
                                        } else {
                                            "???".to_string()
                                        };
                                    ui.add(
                                        Label::new(title)
                                            .with_style(&heading_style_5)
                                            .with_bg_color(
                                                palette
                                                    .color(3)
                                                    .brightness(cymbal_indicator_brightness)
                                                    .into(),
                                            ),
                                    );
                                });
                                ui.add(
                                    ParamKnob::for_param(params.param_float("Cymbal Gain"), setter)
                                        .with_label("Gain")
                                        .with_style(&knob_style_5),
                                );
                                ui.add(
                                    ParamSlider::for_param(
                                        params.param_float("Cymbal Pan"),
                                        setter,
                                    )
                                    .with_label("Pan")
                                    .with_style(&slider_style_5)
                                    .with_indicator_from_center()
                                    .with_orientation(ParamSliderOrientation::Horizontal),
                                );
                                ui.add(
                                    ParamKnob::for_param(
                                        params.param_float("Cymbal Reverb"),
                                        setter,
                                    )
                                    .with_label("Reverb")
                                    .with_style(&knob_style_5),
                                );
                                ui.add(
                                    ParamKnob::for_param(params.param_float("Cymbal Tone"), setter)
                                        .with_label("Filter")
                                        .with_style(&knob_style_5),
                                );
                                ui.add(
                                    ParamKnob::for_param(
                                        params.param_float("Cymbal Decay"),
                                        setter,
                                    )
                                    .with_label("Decay")
                                    .with_style(&knob_style_5),
                                );
                                ui.add_space(bottom_pos - ui.next_widget_position().y); //Push to bottom
                                ui.add(
                                    ParamSlider::for_param(
                                        params.param_int("Cymbal Voice"),
                                        setter,
                                    )
                                    .with_label("Voice")
                                    .with_style(&slider_style_5)
                                    .with_tick_count(3)
                                    .without_alt_tick_length()
                                    .without_fill()
                                    //.with_label_only_value()
                                    .with_length(toggle_length),
                                );
                            });
                            ui.separator();
                            ui.vertical(|ui| {
                                ui.set_width(column_width);
                                ui.vertical_centered(|ui| {
                                    let title =
                                        if let Some(param) = params.param_int("Cowbell Voice") {
                                            param.to_string()
                                        } else {
                                            "???".to_string()
                                        };
                                    ui.add(
                                        Label::new(title)
                                            .with_style(&heading_style_5)
                                            .with_bg_color(
                                                palette
                                                    .color(3)
                                                    .brightness(cowbell_indicator_brightness)
                                                    .into(),
                                            ),
                                    );
                                });
                                ui.add(
                                    ParamKnob::for_param(
                                        params.param_float("Cowbell Gain"),
                                        setter,
                                    )
                                    .with_label("Gain")
                                    .with_style(&knob_style_5),
                                );
                                ui.add(
                                    ParamSlider::for_param(
                                        params.param_float("Cowbell Pan"),
                                        setter,
                                    )
                                    .with_label("Pan")
                                    .with_style(&slider_style_5)
                                    .with_indicator_from_center()
                                    .with_orientation(ParamSliderOrientation::Horizontal),
                                );
                                ui.add(
                                    ParamKnob::for_param(
                                        params.param_float("Cowbell Reverb"),
                                        setter,
                                    )
                                    .with_label("Reverb")
                                    .with_style(&knob_style_5),
                                );
                                ui.add(
                                    ParamKnob::for_param(
                                        params.param_float("Cowbell Transpose"),
                                        setter,
                                    )
                                    .with_label("Tune")
                                    .with_style(&knob_style_5),
                                );
                                ui.add(
                                    ParamKnob::for_param(
                                        params.param_float("Cowbell Alt Transpose"),
                                        setter,
                                    )
                                    .with_label("Alt Tune")
                                    .with_style(&knob_style)
                                    //.with_indicator_from_center(),
                                );
                                ui.add_space(bottom_pos - ui.next_widget_position().y); //Push to bottom
                                ui.add(
                                    ParamSlider::for_param(
                                        params.param_int("Cowbell Voice"),
                                        setter,
                                    )
                                    .with_label("Voice")
                                    .with_style(&slider_style_5)
                                    .with_tick_count(3)
                                    .without_alt_tick_length()
                                    .without_fill()
                                    //.with_label_only_value()
                                    .with_length(toggle_length),
                                );
                            });
                        });
                    });
            },
        )
    }
}

impl ClapPlugin for OneTrickPlugin {
    const CLAP_ID: &'static str = "com.punklabs.onetrick.cryptid";
    const CLAP_DESCRIPTION: Option<&'static str> = Some("An FM Drum Machine");
    const CLAP_MANUAL_URL: Option<&'static str> = Some(Self::URL);
    const CLAP_SUPPORT_URL: Option<&'static str> = None;
    const CLAP_FEATURES: &'static [ClapFeature] = &[
        ClapFeature::Instrument,  //Plugin Category
        ClapFeature::DrumMachine, // Plugin Sub-Category
        ClapFeature::Stereo,
    ]; // Audio Capabilities
       /*
       const CLAP_POLY_MODULATION_CONFIG: Option<PolyModulationConfig> = Some(PolyModulationConfig {
           // If the plugin's voice capacity changes at runtime (for instance, when switching to a
           // monophonic mode), then the plugin should inform the host in the `initialize()` function
           // as well as in the `process()` function if it changes at runtime using
           // `context.set_current_voice_capacity()`
           max_voice_capacity: NUM_VOICES,
           // This enables voice stacking in Bitwig.
           supports_overlapping_voices: true,
       });
       */
}

impl Vst3Plugin for OneTrickPlugin {
    const VST3_CLASS_ID: [u8; 16] = *b"OneTrick-CRYPTID";
    //const VST3_CATEGORIES: &'static str = "Instrument|Synth";
    const VST3_SUBCATEGORIES: &'static [Vst3SubCategory] =
        &[Vst3SubCategory::Instrument, Vst3SubCategory::Synth];
}

nih_export_clap!(OneTrickPlugin);
nih_export_vst3!(OneTrickPlugin);
